import typing

from .encoder import FieldEncoder
from .util import JsonDict

D = typing.TypeVar("D")

DictOrIterable = typing.Union[JsonDict, typing.Iterable[JsonDict]]
DictOrList = typing.Union[JsonDict, typing.List[JsonDict]]

class Serializer(typing.Generic[D]):
    def __init__(
        self,
        dataclass: typing.Type[D],
        omit_none: bool = True,
        type_encoders: typing.Optional[typing.Dict[type, FieldEncoder]] = None,
        only: typing.Optional[typing.List[str]] = None,
        exclude: typing.Optional[typing.List[str]] = None,
        strict: bool = False,
        load_as_type: typing.Optional[type] = None,
    ) -> None: ...
    def json_schema(self, many: bool = False) -> JsonDict: ...
    def get_dict_path(self, obj_path: typing.Sequence[str]) -> typing.List[str]: ...
    def get_object_path(self, dict_path: typing.Sequence[str]) -> typing.List[str]: ...
    @classmethod
    def register_global_type(
        cls,
        field_type: typing.Any,
        encoder: FieldEncoder
    ) -> None: ...
    @classmethod
    def unregister_global_type(cls, field_type: typing.Any) -> None: ...
    def dataclass(self) -> type: ...
    def dump(
        self,
        obj: typing.Union[D, typing.Iterable[D]],
        validate: bool = False,
        many: bool = False,
    ) -> DictOrList: ...
    def load(
        self, data: DictOrIterable, validate: bool = True, many: bool = False
    ) -> typing.Union[D, typing.List[D]]: ...
    def dump_json(
        self,
        obj: typing.Union[D, typing.Iterable[D]],
        validate: bool = False,
        many: bool = False,
    ) -> str: ...
    def load_json(
        self, js: str, validate: bool = True, many: bool = False
    ) -> typing.Union[D, typing.List[D]]: ...



from dataclasses import dataclass

from serpyco import Serializer, post_dump


@dataclass
class Custom(object):
    firstname: str
    lastname: str

    @staticmethod
    @post_dump
    def make_name(data: dict) -> dict:
        first = data["firstname"]
        last = data["lastname"]
        return {"name": f"{first} {last}"}


serializer = Serializer(Custom)
print(serializer.dump(Custom(firstname="foo", lastname="bar")))



import typing

from .util import JsonDict

class FieldEncoder:
    """Base class for encoding fields to and from JSON encodable values"""

    def dump(self, value: typing.Any) -> typing.Any: ...
    def load(self, value: typing.Any) -> typing.Any: ...
    def json_schema(self) -> JsonDict: ...



from dataclasses import dataclass

from serpyco import Serializer, ValidationError, string_field


@dataclass
class StringFields(object):
    simple: str
    name: str = string_field(pattern="^[A-Z]")


serializer = Serializer(StringFields)
print(serializer.load({"name": "Foo", "simple": "whatever"}, validate=True))

try:
    serializer.load({"name": "foo", "simple": "foo"}, validate=True)
except ValidationError as exc:
    print(f"ValidationError: {exc}")



# -*- coding: utf-8 -*-

import abc
import dataclasses
import datetime
import enum
import json
import re
import typing
import uuid
from unittest import mock

import pytest

import serpyco

iso8601_pattern = (
    r"^[0-9]{4}-[0-9][0-9]-[0-9][0-9]T"  # YYYY-MM-DD
    r"[0-9][0-9]:[0-9][0-9]:[0-9][0-9](\.[0-9]+)"  # HH:mm:ss.ssss
    r"?(([+-][0-9][0-9]:[0-9][0-9])|Z)?$"  # timezone
)


class Enum(enum.Enum):
    """
    An enumerate.
    """

    ONE = 1
    TWO = 2


StrType = typing.NewType("StrType", str)


@dataclasses.dataclass
class Simple:
    """
    Basic class.
    """

    name: str


@dataclasses.dataclass
class SimpleWithNewType:
    """
    Basic class.
    """

    name: StrType


@dataclasses.dataclass
class Types:
    """
    Testing class for supported serializer types.
    """

    integer: int
    string: str
    number: float
    boolean: bool
    enum_: Enum
    uid: uuid.UUID
    items: typing.List[str]
    nested: Simple
    nesteds: typing.List[Simple]
    mapping: typing.Dict[str, str]
    datetime_: datetime.datetime
    new_type: StrType
    nested_new_type: SimpleWithNewType
    optional: typing.Optional[int] = None


@pytest.fixture
def types_object() -> Types:
    return Types(
        integer=42,
        string="foo",
        number=12.34,
        boolean=True,
        enum_=Enum.TWO,
        uid=uuid.UUID("12345678123456781234567812345678"),
        items=["one", "two"],
        nested=Simple(name="bar"),
        nesteds=[Simple(name="hello"), Simple(name="world")],
        mapping={"foo": "bar"},
        datetime_=datetime.datetime(2018, 11, 1, 14, 23, 43, 123456),
        new_type=StrType("foo"),
        nested_new_type=SimpleWithNewType(name="bar"),
    )


@dataclasses.dataclass
class First:
    """Circular reference test class"""

    second: "Second"


@dataclasses.dataclass
class Second:
    """Circular reference test class"""

    first: First


@dataclasses.dataclass
class TreeNode:
    """Circular reference test class"""

    sub_nodes: typing.List["TreeNode"]


def test_unit__dump__ok__nominal_case(types_object: Types) -> None:
    serializer = serpyco.Serializer(Types)
    assert {
        "integer": 42,
        "string": "foo",
        "number": 12.34,
        "boolean": True,
        "enum_": 2,
        "uid": "12345678-1234-5678-1234-567812345678",
        "items": ["one", "two"],
        "nested": {"name": "bar"},
        "nesteds": [{"name": "hello"}, {"name": "world"}],
        "mapping": {"foo": "bar"},
        "optional": None,
        "datetime_": "2018-11-01T14:23:43.123456",
        "new_type": "foo",
        "nested_new_type": {"name": "bar"},
    } == serializer.dump(types_object)


def test_unit__dump__ok__with_none(types_object: Types) -> None:
    serializer = serpyco.Serializer(Types, omit_none=False)
    data = serializer.dump(types_object)
    assert "optional" in data
    assert None is data["optional"]


def test_unit__dump_json__ok__nominal_case(types_object: Types) -> None:
    serializer = serpyco.Serializer(Types)

    data = serializer.dump(types_object)
    assert json.dumps(data, separators=(",", ":")) == serializer.dump_json(types_object)


def test_unit__load__ok__nominal_case(types_object: Types) -> None:
    serializer = serpyco.Serializer(Types)
    assert types_object == serializer.load(
        {
            "integer": 42,
            "string": "foo",
            "number": 12.34,
            "boolean": True,
            "enum_": 2,
            "uid": "12345678-1234-5678-1234-567812345678",
            "items": ["one", "two"],
            "nested": {"name": "bar"},
            "nesteds": [{"name": "hello"}, {"name": "world"}],
            "mapping": {"foo": "bar"},
            "datetime_": "2018-11-01T14:23:43.123456",
            "new_type": "foo",
            "nested_new_type": {"name": "bar"},
        }
    )


def test_unit__load_json__ok__nominal_case(types_object: Types) -> None:
    serializer = serpyco.Serializer(Types)
    assert types_object == serializer.load_json(
        json.dumps(
            {
                "integer": 42,
                "string": "foo",
                "number": 12.34,
                "boolean": True,
                "enum_": 2,
                "uid": "12345678-1234-5678-1234-567812345678",
                "items": ["one", "two"],
                "nested": {"name": "bar"},
                "nesteds": [{"name": "hello"}, {"name": "world"}],
                "mapping": {"foo": "bar"},
                "datetime_": "2018-11-01T14:23:43.123456",
                "new_type": "foo",
                "nested_new_type": {"name": "bar"},
            }
        )
    )


def test_unit__from_dump__ok__with_many(types_object: Types) -> None:
    serializer = serpyco.Serializer(Types)

    data = serializer.dump([types_object, types_object], many=True)

    assert [types_object, types_object] == serializer.load(data, many=True)


def test_unit__from_dump_json__ok__with_many(types_object: Types) -> None:
    serializer = serpyco.Serializer(Types)

    data = serializer.dump_json([types_object, types_object], many=True)

    assert [types_object, types_object] == serializer.load_json(data, many=True)


def test_unit__json_schema__ok__nominal_case() -> None:
    serializer = serpyco.Serializer(Types)
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.Types",
        "definitions": {
            "test_unit.Simple": {
                "comment": "test_unit.Simple",
                "description": "Basic class.",
                "properties": {"name": {"type": "string"}},
                "required": ["name"],
                "additionalProperties": True,
                "type": "object",
            },
            "test_unit.SimpleWithNewType": {
                "comment": "test_unit.SimpleWithNewType",
                "description": "Basic class.",
                "properties": {"name": {"type": "string"}},
                "required": ["name"],
                "additionalProperties": True,
                "type": "object",
            },
        },
        "description": "Testing class for supported serializer types.",
        "properties": {
            "boolean": {"type": "boolean"},
            "datetime_": {
                "format": "date-time",
                "type": "string",
                "pattern": iso8601_pattern,
            },
            "enum_": {
                "description": "An enumerate.",
                "enum": [1, 2],
                "type": "integer",
            },
            "uid": {"type": "string", "format": "uuid"},
            "integer": {"type": "integer"},
            "items": {"items": {"type": "string"}, "type": "array"},
            "mapping": {"additionalProperties": {"type": "string"}, "type": "object"},
            "nested": {"$ref": "#/definitions/test_unit.Simple"},
            "nesteds": {
                "items": {"$ref": "#/definitions/test_unit.Simple"},
                "type": "array",
            },
            "number": {"type": "number"},
            "optional": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
            },
            "string": {"type": "string"},
            "new_type": {"type": "string"},
            "nested_new_type": {"$ref": "#/definitions/test_unit.SimpleWithNewType"},
        },
        "required": [
            "integer",
            "string",
            "number",
            "boolean",
            "enum_",
            "uid",
            "items",
            "nested",
            "nesteds",
            "mapping",
            "datetime_",
            "new_type",
            "nested_new_type",
        ],
        "additionalProperties": True,
        "type": "object",
    } == serializer.json_schema(many=False)


def test_unit__json_schema__ok__with_many() -> None:
    serializer = serpyco.Serializer(Types)
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "definitions": {
            "test_unit.Simple": {
                "comment": "test_unit.Simple",
                "description": "Basic class.",
                "properties": {"name": {"type": "string"}},
                "required": ["name"],
                "type": "object",
                "additionalProperties": True,
            },
            "test_unit.SimpleWithNewType": {
                "comment": "test_unit.SimpleWithNewType",
                "description": "Basic class.",
                "properties": {"name": {"type": "string"}},
                "required": ["name"],
                "additionalProperties": True,
                "type": "object",
            },
        },
        "items": {
            "comment": "test_unit.Types",
            "description": "Testing class for supported serializer types.",
            "properties": {
                "boolean": {"type": "boolean"},
                "datetime_": {
                    "format": "date-time",
                    "type": "string",
                    "pattern": iso8601_pattern,
                },
                "enum_": {
                    "description": "An enumerate.",
                    "enum": [1, 2],
                    "type": "integer",
                },
                "uid": {"type": "string", "format": "uuid"},
                "integer": {"type": "integer"},
                "items": {"items": {"type": "string"}, "type": "array"},
                "mapping": {
                    "additionalProperties": {"type": "string"},
                    "type": "object",
                },
                "nested": {"$ref": "#/definitions/test_unit.Simple"},
                "nesteds": {
                    "items": {"$ref": "#/definitions/test_unit.Simple"},
                    "type": "array",
                },
                "number": {"type": "number"},
                "optional": {
                    "anyOf": [{"type": "integer"}, {"type": "null"}],
                    "default": None,
                },
                "string": {"type": "string"},
                "new_type": {"type": "string"},
                "nested_new_type": {
                    "$ref": "#/definitions/test_unit.SimpleWithNewType"
                },
            },
            "required": [
                "integer",
                "string",
                "number",
                "boolean",
                "enum_",
                "uid",
                "items",
                "nested",
                "nesteds",
                "mapping",
                "datetime_",
                "new_type",
                "nested_new_type",
            ],
            "additionalProperties": True,
            "type": "object",
        },
        "type": "array",
    } == serializer.json_schema(many=True)


def test_unit__json_schema__ok__circular_reference() -> None:
    builder = serpyco.SchemaBuilder(First)
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.First",
        "definitions": {
            "test_unit.Second": {
                "comment": "test_unit.Second",
                "description": "Circular reference test class",
                "properties": {"first": {"$ref": "#"}},
                "required": ["first"],
                "type": "object",
                "additionalProperties": True,
            }
        },
        "description": "Circular reference test class",
        "properties": {"second": {"$ref": "#/definitions/test_unit.Second"}},
        "required": ["second"],
        "additionalProperties": True,
        "type": "object",
    } == builder.json_schema()
    nested = builder.nested_builders()
    assert 1 == len(nested)
    assert "test_unit.Second" == nested[0][0]
    assert isinstance(nested[0][1], serpyco.SchemaBuilder)


def test_unit__json_schema__ok__circular_reference_one_class() -> None:
    builder = serpyco.SchemaBuilder(TreeNode)
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.TreeNode",
        "definitions": {},
        "description": "Circular reference test class",
        "properties": {"sub_nodes": {"items": {"$ref": "#"}, "type": "array"}},
        "required": ["sub_nodes"],
        "additionalProperties": True,
        "type": "object",
    } == builder.json_schema()


def test_unit__dump_json__ok__validate() -> None:
    serializer = serpyco.Serializer(Simple)

    assert serializer.dump(Simple(name="foo"), validate=True)
    assert serializer.dump_json(Simple(name="foo"), validate=True)

    with pytest.raises(serpyco.ValidationError):
        serializer.dump(Simple(name=42), validate=True)  # type: ignore
    with pytest.raises(serpyco.ValidationError):
        serializer.dump_json(Simple(name=42), validate=True)  # type: ignore


def test_unit__load_json__ok__validate() -> None:
    serializer = serpyco.Serializer(Simple)

    assert serializer.load({"name": "foo"}, validate=True)
    assert serializer.load_json('{"name": "foo"}', validate=True)

    with pytest.raises(serpyco.ValidationError):
        serializer.load({"name": 42}, validate=True)
    with pytest.raises(serpyco.ValidationError):
        serializer.load_json('{"name": 42}', validate=True)


def test_unit__union__ok__nominal_case() -> None:
    @dataclasses.dataclass
    class WithUnion:
        """Union test class"""

        foo: typing.Union[str, int]

    serializer = serpyco.Serializer(WithUnion)

    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.WithUnion",
        "definitions": {},
        "description": "Union test class",
        "properties": {"foo": {"anyOf": [{"type": "string"}, {"type": "integer"}]}},
        "required": ["foo"],
        "type": "object",
        "additionalProperties": True,
    } == serializer.json_schema()

    assert {"foo": 42} == serializer.dump(WithUnion(foo=42), validate=True)
    assert {"foo": "bar"} == serializer.dump(WithUnion(foo="bar"), validate=True)
    assert WithUnion(foo="bar") == serializer.load({"foo": "bar"})
    with pytest.raises(serpyco.ValidationError):
        serializer.dump(WithUnion(foo=12.34), validate=True)  # type: ignore
    with pytest.raises(serpyco.ValidationError):
        serializer.load({"foo": 12.34})


def test_unit__tuple__ok__nominal_case() -> None:
    @dataclasses.dataclass
    class WithTuple:
        """Tuple test class"""

        tuple_: typing.Tuple[str, int]

    serializer = serpyco.Serializer(WithTuple)
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.WithTuple",
        "definitions": {},
        "description": "Tuple test class",
        "properties": {
            "tuple_": {
                "type": "array",
                "items": [{"type": "string"}, {"type": "integer"}],
                "minItems": 2,
                "maxItems": 2,
            }
        },
        "required": ["tuple_"],
        "additionalProperties": True,
        "type": "object",
    } == serializer.json_schema()
    assert WithTuple(tuple_=("foo", 1)) == serializer.load({"tuple_": ["foo", 1]})


def test_unit__uniform_tuple__ok__nominal_case() -> None:
    @dataclasses.dataclass
    class WithTuple:
        """Tuple test class"""

        tuple_: typing.Tuple[str, ...]

    serializer = serpyco.Serializer(WithTuple)
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.WithTuple",
        "definitions": {},
        "description": "Tuple test class",
        "properties": {"tuple_": {"type": "array", "items": {"type": "string"}}},
        "required": ["tuple_"],
        "type": "object",
        "additionalProperties": True,
    } == serializer.json_schema()
    assert WithTuple(tuple_=("foo", "bar")) == serializer.load(
        {"tuple_": ["foo", "bar"]}
    )


def test_unit__set__ok__nominal_case() -> None:
    @dataclasses.dataclass
    class WithSet:
        """Set test class"""

        set_: typing.Set[str]

    serializer = serpyco.Serializer(WithSet)

    assert WithSet(set_={"foo", "bar"}) == serializer.load({"set_": ["foo", "bar"]})


def test_unit__string_field_format_and_validators__ok__nominal_case() -> None:
    email = mock.Mock()
    datetime_ = mock.Mock()

    @dataclasses.dataclass
    class Nested:
        """Nested"""

        name: str = serpyco.string_field(
            format_=serpyco.StringFormat.DATETIME, validator=datetime_
        )

    @dataclasses.dataclass
    class WithStringField:
        """String field test class"""

        foo: str = serpyco.string_field(
            format_=serpyco.StringFormat.EMAIL,
            validator=email,
            pattern="^[A-Z]",
            min_length=3,
            max_length=24,
        )
        nested: Nested  # type:ignore

    serializer = serpyco.Serializer(WithStringField)

    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.WithStringField",
        "description": "String field test class",
        "definitions": {
            "test_unit.Nested": {
                "comment": "test_unit.Nested",
                "description": "Nested",
                "properties": {"name": {"type": "string", "format": "date-time"}},
                "required": ["name"],
                "type": "object",
                "additionalProperties": True,
            }
        },
        "properties": {
            "foo": {
                "type": "string",
                "format": "email",
                "pattern": "^[A-Z]",
                "minLength": 3,
                "maxLength": 24,
            },
            "nested": {"$ref": "#/definitions/test_unit.Nested"},
        },
        "required": ["foo", "nested"],
        "type": "object",
        "additionalProperties": True,
    } == serializer.json_schema()

    assert serializer.load({"foo": "Foo@foo.bar", "nested": {"name": "foo"}})
    email.assert_called_once_with("Foo@foo.bar")
    datetime_.assert_called_once_with("foo")


def test_unit__number_field__ok__nominal_case() -> None:
    @dataclasses.dataclass
    class WithNumberField:
        """Number field test class"""

        foo: int = serpyco.number_field(minimum=0, maximum=12)

    serializer = serpyco.Serializer(WithNumberField)

    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.WithNumberField",
        "definitions": {},
        "description": "Number field test class",
        "properties": {"foo": {"type": "integer", "minimum": 0, "maximum": 12}},
        "required": ["foo"],
        "type": "object",
        "additionalProperties": True,
    } == serializer.json_schema()

    assert serializer.load({"foo": 5})


def test_unit__field_dict_key__ok__nominal_case() -> None:
    @dataclasses.dataclass
    class WithDictKeyField:
        """Dict key test class"""

        foo: str = serpyco.field(dict_key="bar")

    serializer = serpyco.Serializer(WithDictKeyField)
    assert {"bar": "hello"} == serializer.dump(WithDictKeyField(foo="hello"))
    assert WithDictKeyField(foo="hello") == serializer.load({"bar": "hello"})


def test_unit__type_encoders__ok__nominal_case() -> None:
    class Encoder(serpyco.FieldEncoder):
        def json_schema(self) -> dict:
            return {}

        def dump(self, value):
            return "foo"

    serializer = serpyco.Serializer(Simple, type_encoders={str: Encoder()})

    assert {"name": "foo"} == serializer.dump(Simple(name="bar"))
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.Simple",
        "description": "Basic class.",
        "definitions": {},
        "properties": {"name": {}},
        "required": ["name"],
        "type": "object",
        "additionalProperties": True,
    } == serializer.json_schema()

    second = serpyco.Serializer(Simple)
    assert {"name": "bar"} == second.dump(Simple(name="bar"))
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.Simple",
        "description": "Basic class.",
        "definitions": {},
        "properties": {"name": {"type": "string"}},
        "required": ["name"],
        "type": "object",
        "additionalProperties": True,
    } == second.json_schema()

    @dataclasses.dataclass
    class Nest:
        """Nest"""

        name: str
        nested: Simple = serpyco.nested_field(type_encoders={str: Encoder()})

    serializer = serpyco.Serializer(Nest)
    assert {"name": "bar", "nested": {"name": "foo"}} == serializer.dump(
        Nest(name="bar", nested=Simple("bar"))
    )
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.Nest",
        "description": "Nest",
        "definitions": {
            "test_unit.Simple": {
                "comment": "test_unit.Simple",
                "description": "Basic class.",
                "properties": {"name": {}},
                "required": ["name"],
                "type": "object",
                "additionalProperties": True,
            }
        },
        "properties": {
            "name": {"type": "string"},
            "nested": {"$ref": "#/definitions/test_unit.Simple"},
        },
        "required": ["name", "nested"],
        "type": "object",
        "additionalProperties": True,
    } == serializer.json_schema()


def test_unit__global_type_encoders__ok__nominal_case() -> None:
    class Encoder(serpyco.FieldEncoder):
        def json_schema(self) -> dict:
            return {}

        def dump(self, value):
            return "foo"

    serpyco.Serializer.register_global_type(str, Encoder())

    serializer = serpyco.Serializer(Simple)

    assert {"name": "foo"} == serializer.dump(Simple(name="bar"))
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.Simple",
        "description": "Basic class.",
        "definitions": {},
        "properties": {"name": {}},
        "required": ["name"],
        "type": "object",
        "additionalProperties": True,
    } == serializer.json_schema()

    second = serpyco.Serializer(Simple)
    assert {"name": "foo"} == second.dump(Simple(name="bar"))
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.Simple",
        "description": "Basic class.",
        "definitions": {},
        "properties": {"name": {}},
        "required": ["name"],
        "type": "object",
        "additionalProperties": True,
    } == second.json_schema()

    serpyco.Serializer.unregister_global_type(str)

    third = serpyco.Serializer(Simple)
    assert {"name": "bar"} == third.dump(Simple(name="bar"))
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.Simple",
        "description": "Basic class.",
        "definitions": {},
        "properties": {"name": {"type": "string"}},
        "required": ["name"],
        "type": "object",
        "additionalProperties": True,
    } == third.json_schema()


def test_unit__global_type_encoders_metaclass__ok__no_exception() -> None:
    class WithMeta(metaclass=abc.ABCMeta):
        pass

    class Encoder(serpyco.FieldEncoder):
        def dump(self, my: WithMeta):
            return None

        def json_schema(self) -> dict:
            return {}

    serpyco.Serializer.register_global_type(WithMeta, Encoder())
    serpyco.Serializer.unregister_global_type(WithMeta)


def test_unit__ignore__ok__nominal_case():
    @dataclasses.dataclass
    class Ignore:
        """Ignore test class"""

        foo: str = serpyco.field(ignore=True)

    serializer = serpyco.Serializer(Ignore)
    assert {} == serializer.dump(Ignore(foo="bar"))
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.Ignore",
        "description": "Ignore test class",
        "definitions": {},
        "properties": {},
        "type": "object",
        "additionalProperties": True,
        "required": [],
    } == serializer.json_schema()


def test_unit__only__ok__nominal_case():
    @dataclasses.dataclass
    class Only:
        """Only test class"""

        foo: str
        bar: str

    serializer = serpyco.Serializer(Only, only=["foo"])
    assert {"foo": "bar"} == serializer.dump(Only(foo="bar", bar="foo"))
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.Only",
        "description": "Only test class",
        "definitions": {},
        "properties": {"foo": {"type": "string"}},
        "required": ["foo"],
        "type": "object",
        "additionalProperties": True,
    } == serializer.json_schema()


def test_unit__field_description_and_examples__ok__nominal_case():
    @dataclasses.dataclass
    class Desc:
        """Description test class"""

        foo: str = serpyco.field(
            description="This is a foo", examples=["can be foo", "or bar"]
        )

    serializer = serpyco.Serializer(Desc)
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.Desc",
        "definitions": {},
        "description": "Description test class",
        "properties": {
            "foo": {
                "description": "This is a foo",
                "examples": ["can be foo", "or bar"],
                "type": "string",
            }
        },
        "required": ["foo"],
        "type": "object",
        "additionalProperties": True,
    } == serializer.json_schema()


def test_unit__field_default__ok__nominal_case():
    @dataclasses.dataclass
    class Desc:
        """Description test class"""

        foo: str = "foo"
        bar: str = dataclasses.field(default_factory=lambda: "bar")
        datetime_: datetime.datetime = datetime.datetime(2018, 11, 24, 19, 0, 0, 0)

    serializer = serpyco.Serializer(Desc)
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.Desc",
        "definitions": {},
        "description": "Description test class",
        "properties": {
            "foo": {"default": "foo", "type": "string"},
            "bar": {"type": "string"},
            "datetime_": {
                "default": "2018-11-24T19:00:00",
                "type": "string",
                "format": "date-time",
                "pattern": iso8601_pattern,
            },
        },
        "type": "object",
        "additionalProperties": True,
        "required": [],
    } == serializer.json_schema()


def test_unit__decorators__ok__nominal_case():
    @dataclasses.dataclass
    class Decorated:
        foo: typing.Optional[str]
        bar: int

        @staticmethod
        @serpyco.pre_dump
        def add_two_to_bar(obj: "Decorated") -> "Decorated":
            obj.bar += 2
            return obj

        @staticmethod
        @serpyco.post_dump
        def del_foo_key(data: dict) -> dict:
            del data["foo"]
            return data

        @staticmethod
        @serpyco.pre_load
        def add_foo_if_missing(data: dict) -> dict:
            if "foo" not in data:
                data["foo"] = "default"
            return data

        @staticmethod
        @serpyco.post_load
        def substract_two_from_bar(obj: "Decorated") -> "Decorated":
            obj.bar -= 2
            return obj

    serializer = serpyco.Serializer(Decorated)

    assert {"bar": 5} == serializer.dump(Decorated(foo="hello", bar=3))

    assert '{"bar":5}' == serializer.dump_json(Decorated(foo="hello", bar=3))

    assert Decorated(foo="default", bar=1) == serializer.load({"bar": 3})

    assert Decorated(foo="default", bar=1) == serializer.load_json('{"bar":3}')


def test_unit__exclude__ok__nominal_case():
    @dataclasses.dataclass
    class Exclude:
        """Exclude test class"""

        foo: str
        bar: str

    serializer = serpyco.Serializer(Exclude, exclude=["foo"])
    assert {"bar": "foo"} == serializer.dump(Exclude(foo="bar", bar="foo"))
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.Exclude",
        "description": "Exclude test class",
        "definitions": {},
        "properties": {"bar": {"type": "string"}},
        "required": ["bar"],
        "type": "object",
        "additionalProperties": True,
    } == serializer.json_schema()


def test_unit__nested_field__ok__nominal_case():
    @dataclasses.dataclass
    class Nested:
        """Nested test class"""

        foo: str
        bar: str

    @dataclasses.dataclass
    class Parent:
        """Parent test class"""

        first: Nested = serpyco.nested_field(only=["foo"])
        second: Nested = serpyco.nested_field(exclude=["foo"])

    serializer = serpyco.Serializer(Parent)
    assert {"first": {"foo": "foo"}, "second": {"bar": "bar"}} == serializer.dump(
        Parent(first=Nested(foo="foo", bar="bar"), second=Nested(foo="foo", bar="bar"))
    )
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.Parent",
        "description": "Parent test class",
        "definitions": {
            "test_unit.Nested_exclude_foo": {
                "type": "object",
                "additionalProperties": True,
                "comment": "test_unit.Nested",
                "description": "Nested test class",
                "properties": {"bar": {"type": "string"}},
                "required": ["bar"],
            },
            "test_unit.Nested_exclude_bar": {
                "type": "object",
                "additionalProperties": True,
                "comment": "test_unit.Nested",
                "description": "Nested test class",
                "properties": {"foo": {"type": "string"}},
                "required": ["foo"],
            },
        },
        "properties": {
            "first": {"$ref": "#/definitions/test_unit.Nested_exclude_bar"},
            "second": {"$ref": "#/definitions/test_unit.Nested_exclude_foo"},
        },
        "required": ["first", "second"],
        "type": "object",
        "additionalProperties": True,
    } == serializer.json_schema()


def test_unit__get_dict_object_path__ok__nominal_case():
    @dataclasses.dataclass
    class Nested:
        foo: str = serpyco.field(dict_key="bar")

    @dataclasses.dataclass
    class Parent:
        nested: Nested = serpyco.field(dict_key="n")
        nesteds: typing.List[Nested] = serpyco.field(dict_key="ns")
        mapped: typing.Dict[str, Nested] = serpyco.field(dict_key="mp")

    serializer = serpyco.Serializer(Parent)

    assert ["n", "bar"] == serializer.get_dict_path(["nested", "foo"])
    assert ["nested", "foo"] == serializer.get_object_path(["n", "bar"])
    assert ["ns", "bar"] == serializer.get_dict_path(["nesteds", "foo"])
    assert ["nesteds", "foo"] == serializer.get_object_path(["ns", "bar"])
    assert ["mp", "bar"] == serializer.get_dict_path(["mapped", "foo"])
    assert ["mapped", "foo"] == serializer.get_object_path(["mp", "bar"])


def test_unit__dict_encoder__ok__nominal_case():
    class CustomEncoder(serpyco.FieldEncoder):
        def dump(self, value):
            return value

        def load(self, value):
            return value

    @dataclasses.dataclass
    class Nested:
        foo: str

    @dataclasses.dataclass
    class Parent:
        mapping: typing.Dict[str, Nested]
        custom: typing.Dict[int, Nested]

    serializer = serpyco.Serializer(Parent, type_encoders={int: CustomEncoder()})
    assert {
        "mapping": {"foo": {"foo": "bar"}},
        "custom": {42: {"foo": "foo"}},
    } == serializer.dump(
        Parent(mapping={"foo": Nested(foo="bar")}, custom={42: Nested(foo="foo")})
    )


def test_unit__rapidjson_validator__err_message():
    @dataclasses.dataclass
    class Foo:
        name: str

    val = serpyco.validator.RapidJsonValidator(serpyco.SchemaBuilder(Foo))
    with pytest.raises(
        serpyco.ValidationError,
        match=r'value "42" at path "#/name" has type "int", expected "string"',
    ):
        val.validate({"name": 42})


def test_unit__field_cast_on_load__ok__nominal_case():
    @dataclasses.dataclass
    class CastOnLoad:
        value: int = serpyco.field(cast_on_load=True)

    serializer = serpyco.Serializer(CastOnLoad)
    assert CastOnLoad(value=42) == serializer.load({"value": "42"})


def test_unit__field_cast_on_load__err_exception_during_casting():
    @dataclasses.dataclass
    class CastOnLoad:
        value: int = serpyco.field(cast_on_load=True)

    serializer = serpyco.Serializer(CastOnLoad)
    with pytest.raises(serpyco.ValidationError):
        serializer.load({"value": "hello"})


def test_unit__custom_definition_name__ok__nominal_case():
    @dataclasses.dataclass
    class Nested:
        """Nested"""

        value: int

    @dataclasses.dataclass
    class Class:
        """Class"""

        nested: Nested

    get_definition_name = mock.Mock(return_value="Custom")
    builder = serpyco.SchemaBuilder(Class, get_definition_name=get_definition_name)

    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.Class",
        "definitions": {
            "Custom": {
                "comment": "test_unit.Nested",
                "description": "Nested",
                "properties": {"value": {"type": "integer"}},
                "required": ["value"],
                "type": "object",
                "additionalProperties": True,
            }
        },
        "description": "Class",
        "properties": {"nested": {"$ref": "#/definitions/Custom"}},
        "required": ["nested"],
        "type": "object",
        "additionalProperties": True,
    } == builder.json_schema()
    get_definition_name.assert_called_with(Nested, (), [])


def test_unit__not_init_fields__ok__nominal_case():
    @dataclasses.dataclass
    class Class:
        """Class"""

        one: str
        two: int = dataclasses.field(init=False)

    serializer = serpyco.Serializer(Class)
    obj = Class(one="hello")
    obj.two = 12
    assert obj == serializer.load({"one": "hello", "two": 12})
    assert {"one": "hello", "two": 12} == serializer.dump(obj)


def test_unit__schema__ok__with_default_dataclass():
    @dataclasses.dataclass
    class Nested:
        """Nested"""

        name: str = "Hello"

    @dataclasses.dataclass
    class Class:
        """Class"""

        one: str
        nested: Nested = dataclasses.field(default_factory=Nested)

    serializer = serpyco.Serializer(Class)
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.Class",
        "definitions": {
            "test_unit.Nested": {
                "comment": "test_unit.Nested",
                "description": "Nested",
                "properties": {"name": {"default": "Hello", "type": "string"}},
                "type": "object",
                "additionalProperties": True,
                "required": [],
            }
        },
        "description": "Class",
        "properties": {
            "nested": {"$ref": "#/definitions/test_unit.Nested"},
            "one": {"type": "string"},
        },
        "required": ["one"],
        "type": "object",
        "additionalProperties": True,
    } == serializer.json_schema()


def test_unit__generic_dataclass__ok__nominal_case():
    T = typing.TypeVar("T")

    @dataclasses.dataclass
    class Gen(typing.Generic[T]):
        "Generic."
        foo: str
        bar: T

    serializer = serpyco.Serializer(Gen[int])
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.Gen",
        "definitions": {},
        "description": "Generic.",
        "properties": {"bar": {"type": "integer"}, "foo": {"type": "string"}},
        "required": ["foo", "bar"],
        "type": "object",
        "additionalProperties": True,
    } == serializer.json_schema()

    @dataclasses.dataclass
    class WithGen:
        "With a generic."
        nested: Gen[int]

    serializer = serpyco.Serializer(WithGen)
    assert {
        "comment": "test_unit.WithGen",
        "$schema": "http://json-schema.org/draft-04/schema#",
        "definitions": {
            "test_unit.Gen[int]": {
                "comment": "test_unit.Gen",
                "description": "Generic.",
                "properties": {"bar": {"type": "integer"}, "foo": {"type": "string"}},
                "required": ["foo", "bar"],
                "type": "object",
                "additionalProperties": True,
            }
        },
        "description": "With a generic.",
        "properties": {"nested": {"$ref": "#/definitions/test_unit.Gen[int]"}},
        "required": ["nested"],
        "type": "object",
        "additionalProperties": True,
    } == serializer.json_schema()
    assert WithGen(nested=Gen(foo="bar", bar=12)) == serializer.load(
        {"nested": {"foo": "bar", "bar": 12}}
    )

    @dataclasses.dataclass(init=False)
    class SList(typing.Generic[T]):
        """List."""

        items: typing.List[T]
        item_nb: int

        def __init__(self, items: typing.List[T]) -> None:
            self.items = items
            self.item_nb = len(items)

    serializer = serpyco.Serializer(SList[int])
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.SList",
        "definitions": {},
        "description": "List.",
        "properties": {
            "item_nb": {"type": "integer"},
            "items": {"items": {"type": "integer"}, "type": "array"},
        },
        "required": ["items", "item_nb"],
        "type": "object",
        "additionalProperties": True,
    } == serializer.json_schema()
    assert {"items": [0, 1], "item_nb": 2} == serializer.dump(SList([0, 1]))


def test_unit__schema__ok__none_default():
    @dataclasses.dataclass
    class Def:
        """Def."""

        foo: typing.Optional[int] = None

    serializer = serpyco.Serializer(Def)
    assert {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.Def",
        "definitions": {},
        "description": "Def.",
        "properties": {
            "foo": {"anyOf": [{"type": "integer"}, {"type": "null"}], "default": None}
        },
        "type": "object",
        "required": [],
        "additionalProperties": True,
    } == serializer.json_schema()


def test_unit__union_field_encoder__ok__nominal_case():

    dummy = mock.Mock()
    dummy.load.side_effect = lambda v: int(v)
    dummy_raise_at_load = mock.Mock()
    dummy_raise_at_load.load.side_effect = Exception
    encoder = serpyco.serializer.UnionFieldEncoder(
        [(int, dummy), (str, dummy_raise_at_load)]
    )
    encoder.dump(42)
    dummy.dump.assert_called_once_with(42)
    dummy_raise_at_load.dump.assert_not_called

    encoder.load(42)
    dummy.load.assert_called_once_with(42)
    dummy_raise_at_load.load.assert_not_called

    dummy.load.reset_mock()
    with pytest.raises(serpyco.ValidationError):
        encoder.load("hello")
        dummy.load.assert_called_once_with("hello")
        dummy_raise_at_load.load.assert_called_once_with("hello")


def test_unit__union_field_encoder__err__validation_error():

    dummy_raise_at_load = mock.Mock()
    dummy_raise_at_load.load.side_effect = Exception
    encoder = serpyco.serializer.UnionFieldEncoder(
        [(int, dummy_raise_at_load), (str, dummy_raise_at_load)]
    )
    with pytest.raises(serpyco.ValidationError):
        encoder.load("hello")


def test_unit__optional__custom_encoder__ok__nominal_case():
    @dataclasses.dataclass
    class OptionalCustom:
        """OptionalCustom."""

        name: typing.Optional[str]

    class CustomEncoder(serpyco.FieldEncoder):
        def json_schema(self):
            return {"type": "string"}

    serializer = serpyco.Serializer(
        OptionalCustom, type_encoders={str: CustomEncoder()}
    )

    assert serializer.json_schema() == {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.OptionalCustom",
        "definitions": {},
        "description": "OptionalCustom.",
        "properties": {"name": {"anyOf": [{"type": "string"}, {"type": "null"}]}},
        "type": "object",
        "additionalProperties": True,
        "required": ["name"],
    }


def test_unit__serializer__err__nested_not_dataclass():
    class Nested:
        pass

    @dataclasses.dataclass
    class Foo:
        n: Nested

    with pytest.raises(serpyco.NoEncoderError):
        serpyco.Serializer(Foo)

    with pytest.raises(serpyco.SchemaError):
        b = serpyco.SchemaBuilder(Foo)
        b._create_json_schema()


def test_unit__schema__ok__allowed_values():
    @dataclasses.dataclass
    class WithAllowedValues:
        """WithAllowedValues."""

        foo: int = serpyco.number_field(allowed_values=[1, 2, 3])

    serializer = serpyco.Serializer(WithAllowedValues)

    assert serializer.json_schema() == {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.WithAllowedValues",
        "definitions": {},
        "description": "WithAllowedValues.",
        "properties": {"foo": {"type": "integer", "enum": [1, 2, 3]}},
        "required": ["foo"],
        "type": "object",
        "additionalProperties": True,
    }


def test_unit__schema__ok__allowed_values_with_enum():
    class Enumerate(enum.Enum):
        """Enum."""

        ONE = 1
        TWO = 2
        THREE = 3

    @dataclasses.dataclass
    class WithAllowedValues:
        """WithAllowedValues."""

        foo: Enumerate = serpyco.field(allowed_values=[Enumerate.ONE, Enumerate.TWO])

    serializer = serpyco.Serializer(WithAllowedValues)

    assert serializer.json_schema() == {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "comment": "test_unit.WithAllowedValues",
        "definitions": {},
        "description": "WithAllowedValues.",
        "properties": {
            "foo": {"type": "integer", "enum": [1, 2], "description": "Enum."}
        },
        "required": ["foo"],
        "type": "object",
        "additionalProperties": True,
    }


def test_unit__cast_on_load__ok__with_optional():
    @dataclasses.dataclass
    class OptionalCastOnLoad:
        foo: typing.Optional[int] = serpyco.field(cast_on_load=True)

    serializer = serpyco.Serializer(OptionalCastOnLoad)

    assert OptionalCastOnLoad(foo=2) == serializer.load({"foo": "2"})


def test_unit__dict__ok__with_field_encoder():
    class NeedFieldEncoder(str, enum.Enum):
        FOO = "FOO"
        BAR = "BAR"

    @dataclasses.dataclass
    class WithFieldEncoder:
        foo: typing.Dict[str, NeedFieldEncoder]
        bar: typing.Dict[NeedFieldEncoder, str]
        foobar: typing.Dict[NeedFieldEncoder, NeedFieldEncoder]

    serializer = serpyco.Serializer(WithFieldEncoder)

    obj = WithFieldEncoder(
        foo={"f": NeedFieldEncoder.FOO},
        bar={NeedFieldEncoder.BAR: "b"},
        foobar={NeedFieldEncoder.FOO: NeedFieldEncoder.BAR},
    )

    dict_ = {"foo": {"f": "FOO"}, "bar": {"BAR": "b"}, "foobar": {"FOO": "BAR"}}

    assert serializer.dump(obj) == dict_

    assert serializer.load(dict_) == obj


def test_unit__optional__ok__with_validator():
    validate = mock.Mock()

    @dataclasses.dataclass
    class WithVal:
        foo: typing.Optional[str] = serpyco.field(validator=validate, default="bar")

    serializer = serpyco.Serializer(WithVal)

    assert serializer.load({}) == WithVal()
    validate.assert_not_called()


def test_unit__optional__err__validation_error_message():
    @dataclasses.dataclass
    class Opt:
        foo: typing.Optional[str] = serpyco.string_field(min_length=5)

    serializer = serpyco.Serializer(Opt)

    with pytest.raises(
        serpyco.ValidationError,
        match=re.escape(
            r'value "bar" at path "#/foo" must have its length >= 5 but length is 3'
        ),
    ):
        serializer.load({"foo": "bar"})


def test_unit__embedded_dataclass_list__ok__with_validator():

    validator = mock.Mock()

    @dataclasses.dataclass
    class Foo:
        bar: str = serpyco.field(validator=validator)

    @dataclasses.dataclass
    class ListFoo:
        foos: typing.List[Foo]

    serializer = serpyco.Serializer(ListFoo)
    serializer.load({"foos": [{"bar": "hello"}, {"bar": "world"}]})
    assert 2 == validator.call_count


def test_unit__validation__ok__several_errors():
    @dataclasses.dataclass
    class Foo:
        bar: str
        foo: int

    serializer = serpyco.Serializer(Foo)
    with pytest.raises(
        serpyco.ValidationError,
        match=re.escape(
            r'- value "12" at path "#/bar" has type "int", expected "string"\n'
            r'- value "hello" at path "#/foo" has type "str", expected "integer"'
        ),
    ):
        serializer.load({"bar": 12, "foo": "hello"})


def test_unit__validation__ok__empty_content():
    @dataclasses.dataclass
    class Foo:
        bar: str

    with pytest.raises(serpyco.ValidationError):
        serpyco.Serializer(Foo).load({})


def test_unit__validation_error_message__err__optional_sub_dataclass():
    @dataclasses.dataclass
    class Foo:
        bar: str

    @dataclasses.dataclass
    class Bar:
        hello: int
        foo: typing.Optional[Foo]

    with pytest.raises(
        serpyco.ValidationError,
        match=re.escape(r'value "{}" at path "#/foo" must define property "bar"'),
    ):
        serpyco.Serializer(Bar).load({"hello": 42, "foo": {}})


def test_unit__strict_validation__err__additional_property():
    @dataclasses.dataclass
    class Foo:
        bar: str

    serializer = serpyco.Serializer(Foo, strict=True)
    with pytest.raises(
        serpyco.ValidationError,
        match=re.escape(r'properties "hello" cannot be defined'),
    ):
        serializer.load({"hello": 42, "bar": "foo"})

    assert serializer.load({"bar": "foo"}) == Foo("foo")


def test_unit__dump_load__object_set():
    @dataclasses.dataclass
    class Bar:
        baz: str

        def __hash__(self):
            return hash(self.baz)

    @dataclasses.dataclass
    class Foo:
        bar: typing.Set[Bar]

    serializer = serpyco.Serializer(Foo)
    foo = Foo({Bar("baz")})
    assert serializer.dump(foo) == {"bar": [{"baz": "baz"}]}
    assert serializer.load({"bar": [{"baz": "baz"}]}) == foo


def test_unit__load__err__missing_parameter():
    @dataclasses.dataclass
    class Foo:
        name: str
        value: int

    serializer = serpyco.Serializer(Foo)
    with pytest.raises(TypeError):
        serializer.load({"name": "hello"}, validate=False)


def test_unit__load__ok__custom_type():
    class Bar:
        def __init__(self, id):
            self.id = id

        def __eq__(self, other):
            return isinstance(other, Bar) and other.id == self.id

    class Foo:
        def __init__(self, name, bar):
            self.name = name
            self.bar = bar

        def __eq__(self, other):
            return (
                isinstance(other, Foo)
                and other.name == self.name
                and other.bar == self.bar
            )

    @dataclasses.dataclass
    class BarSchema:
        id: int

    @dataclasses.dataclass
    class FooSchema:
        name: str
        bar: BarSchema = serpyco.nested_field(load_as_type=Bar)

    serializer = serpyco.Serializer(FooSchema, load_as_type=Foo)

    assert Foo("hello", Bar(1)) == serializer.load({"name": "hello", "bar": {"id": 1}})


def test_unit__load__ok__frozen_dataclass():
    @dataclasses.dataclass(frozen=True)
    class Frozen:
        name: str

    serializer = serpyco.Serializer(Frozen)

    assert Frozen("hello") == serializer.load({"name": "hello"})


def test_unit__ok__schema_optional_dict_dataclass():
    @dataclasses.dataclass
    class Foo:
        name: str

    @dataclasses.dataclass
    class FooDict:
        foos: typing.Optional[typing.Dict[str, Foo]]

    @dataclasses.dataclass
    class Bar:
        foo_dict: FooDict

    schema = serpyco.SchemaBuilder(Bar).json_schema()
    assert "test_unit.Foo" in list(schema["definitions"].keys())


def test_unit_load__ok__excluded_fields():
    @dataclasses.dataclass
    class Foo:
        name: str
        description: str = ""

    serializer = serpyco.Serializer(Foo, exclude=["description"])
    assert Foo(name="foo") == serializer.load(
        {"name": "foo", "description": "excluded"}
    )

    serializer = serpyco.Serializer(Foo, only=["name"])
    assert Foo(name="foo") == serializer.load(
        {"name": "foo", "description": "excluded"}
    )

    @dataclasses.dataclass
    class Bar:
        name: str
        description: str = serpyco.field(ignore=True, default="")

    serializer = serpyco.Serializer(Bar)
    assert Bar(name="foo") == serializer.load(
        {"name": "foo", "description": "excluded"}
    )


def test_unit_load__err__excluded_fields_without_default():
    @dataclasses.dataclass
    class Foo:
        name: str
        description: str

    serializer = serpyco.Serializer(Foo, exclude=["description"])
    with pytest.raises(TypeError):
        serializer.load({"name": "foo", "description": "excluded"})

    serializer = serpyco.Serializer(Foo, only=["name"])
    with pytest.raises(TypeError):
        serializer.load({"name": "foo", "description": "excluded"})

    @dataclasses.dataclass
    class Bar:
        name: str
        description: str = serpyco.field(ignore=True)

    serializer = serpyco.Serializer(Bar)
    with pytest.raises(TypeError):
        serializer.load({"name": "foo", "description": "excluded"})


def test_unit__ok__serializer_mixin():
    @dataclasses.dataclass
    class Foo(serpyco.SerializerMixin):
        name: str

    assert {"name": "Hello"} == Foo(name="Hello").dump()
    assert Foo(name="Hello") == Foo.load({"name": "Hello"})
    assert '{"name":"Hello"}' == Foo(name="Hello").dump_json()
    assert Foo(name="Hello") == Foo.load_json('{"name":"Hello"}')


def test_unit_json_schema__ok__ignore_sub_field():
    @dataclasses.dataclass
    class Bar:
        """Bar"""

        value: int
        comment: str = serpyco.field(ignore=True)

    @dataclasses.dataclass
    class Foo:
        """Foo"""

        name: str
        bar: Bar

    schema_builder = serpyco.SchemaBuilder(Foo)
    assert schema_builder.json_schema() == {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "additionalProperties": True,
        "comment": "test_unit.Foo",
        "definitions": {
            "test_unit.Bar_exclude_comment": {
                "additionalProperties": True,
                "comment": "test_unit.Bar",
                "description": "Bar",
                "properties": {"value": {"type": "integer"}},
                "required": ["value"],
                "type": "object",
            }
        },
        "description": "Foo",
        "properties": {
            "bar": {"$ref": "#/definitions/test_unit.Bar_exclude_comment"},
            "name": {"type": "string"},
        },
        "required": ["name", "bar"],
        "type": "object",
    }


def test_unit_load__ok__post_init():
    @dataclasses.dataclass
    class Foo:
        def __post_init__(self) -> None:
            self.name = "post_init_called"

        name: str

    foo = serpyco.Serializer(Foo).load({"name": "not_called"})
    assert "post_init_called" == foo.name


def test_unit_serializer__ok__ignored_not_dataclass_field():
    class Foo:
        def __init__(self) -> None:
            self.name = "Foo"

    @dataclasses.dataclass
    class Bar:
        value: int
        foo: Foo = serpyco.field(ignore=True)

    serpyco.Serializer(Bar)


def test_unit_field_description__ok__nominal_case():
    @dataclasses.dataclass
    class Foo:
        value: int = serpyco.field(description="An integer")

    schema = serpyco.SchemaBuilder(Foo).json_schema()
    assert "An integer" == schema["properties"]["value"]["description"]


def test_unit_optional_dataclass_list__ok__nominal_case():
    @dataclasses.dataclass
    class SubTest:
        value: str

    @dataclasses.dataclass
    class Test:
        value: typing.List[typing.Optional[SubTest]]

    serializer = serpyco.Serializer(Test)
    assert serializer.dump(Test([None])) == {"value": [None]}


def test_unit_union_with_none__ok__nominal_case():
    @dataclasses.dataclass
    class SubTest1:
        value1: str

    @dataclasses.dataclass
    class SubTest2:
        value2: str

    @dataclasses.dataclass
    class Test:
        value: typing.List[typing.Optional[typing.Union[SubTest1, SubTest2]]]

    serializer = serpyco.Serializer(Test)
    assert serializer.dump(Test(value=[None, SubTest1("1"), SubTest2("2")])) == {
        "value": [
            None,
            {"value1": "1"},
            {"value2": "2"},
        ]
    }


def test_unit_untyped_collections__ok__nominal_case():
    @dataclasses.dataclass
    class Untyped:
        """Untyped collection fields"""

        udict: dict
        tdict: typing.Dict
        ulist: list
        tlist: typing.List
        uset: set
        tset: typing.Set

    serializer = serpyco.Serializer(Untyped)
    assert serializer.dump(
        Untyped(
            udict={}, tdict={"foo": 1}, ulist=[], tlist=[0, 2], uset=set(), tset={0, 12}
        )
    ) == {
        "udict": {},
        "tdict": {"foo": 1},
        "ulist": [],
        "tlist": [0, 2],
        "uset": [],
        "tset": [0, 12],
    }

    assert serializer.json_schema() == {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "additionalProperties": True,
        "comment": "test_unit.Untyped",
        "definitions": {},
        "description": "Untyped collection fields",
        "properties": {
            "tdict": {"additionalProperties": {}, "type": "object"},
            "tlist": {"additionalProperties": {}, "type": "array"},
            "tset": {"additionalProperties": {}, "type": "array"},
            "udict": {"additionalProperties": {}, "type": "object"},
            "ulist": {"additionalProperties": {}, "type": "array"},
            "uset": {"additionalProperties": {}, "type": "array"},
        },
        "required": ["udict", "tdict", "ulist", "tlist", "uset", "tset"],
        "type": "object",
    }


def test_unit_bytes_as_field__err__nominal_case():
    @dataclasses.dataclass
    class WithBytes:
        wrong: bytes

    with pytest.raises(serpyco.NoEncoderError):
        serpyco.Serializer(WithBytes)


def test_unit_bytes_input__err__nominal_case():
    @dataclasses.dataclass
    class Foo:
        wrong: str

    serializer = serpyco.Serializer(Foo)
    with pytest.raises(serpyco.ValidationError):
        serializer.load({"wrong": b"abc"}, validate=True)



# -*- coding: utf-8 -*-
import enum
import typing

_serpyco_tags = "__serpyco_tags__"


class DecoratorType(str, enum.Enum):
    PRE_DUMP = "pre_dump"
    POST_DUMP = "post_dump"
    PRE_LOAD = "pre_load"
    POST_LOAD = "post_load"


ObjCallable = typing.Callable[[object], object]
DictCallable = typing.Callable[
    [typing.Dict[str, typing.Any]], typing.Dict[str, typing.Any]
]


def pre_dump(method: ObjCallable) -> ObjCallable:
    """
    This decorator can be applied to a callable taking one object
    and should return an object of the dataclass it is declared in.
    The method will then be called with each object given to Serializer.dump
    or Serializer.dump_json before dumping them.
    :param: method method to call before dumping
    """
    setattr(method, _serpyco_tags, DecoratorType.PRE_DUMP)
    return method


def post_load(method: ObjCallable) -> ObjCallable:
    """
    This decorator can be applied to a callable taking one data class object
    and should return an object.
    The method will then be called with each object output by Serializer.load
    or Serializer.load_json before returning them.
    :param: method method to call after loading.
    """
    setattr(method, _serpyco_tags, DecoratorType.POST_LOAD)
    return method


def pre_load(method: DictCallable) -> DictCallable:
    """
    This decorator can be applied to a callable taking one dictionary
    and should return another dictionary.
    The method will then be called with each dictionary given to
    Serializer.load or Serializer.load before loading them in
    dataclass objects.
    :param: method method to call before loading
    """
    setattr(method, _serpyco_tags, DecoratorType.PRE_LOAD)
    return method


def post_dump(method: DictCallable) -> DictCallable:
    """
    This decorator can be applied to a callable taking one dictionary
    and should return another dictionary.
    The method will then be called with each dictionary output by
    Serializer.dump or Serializer.dump_json after dumping them.
    :param: method method to call after dumping
    """
    setattr(method, _serpyco_tags, DecoratorType.POST_DUMP)
    return method



# -*- coding: utf-8 -*-

from dataclasses import dataclass
from pprint import pprint

from serpyco import Serializer


@dataclass
class Point(object):
    x: float
    y: float


serializer = Serializer(Point)

pprint(serializer.json_schema())
pprint(serializer.load({"x": 3.14, "y": 1.5}))
try:
    serializer.load({"x": 3.14, "y": "wrong"})
except Exception as ex:
    pprint(ex)
pprint(serializer.dump(Point(x=3.14, y=1.5)))
try:
    serializer.dump(Point(x=3.14, y="wrong"), validate=True)
except Exception as ex:
    pprint(ex)



# -*- coding: utf-8 -*-

from setuptools import Extension, find_packages, setup

requires = [
    "python-dateutil",
    "python-rapidjson",
    "typing-inspect>=0.6.0",
    "dataclasses;python_version<'3.7'",
]

with open("README.rst") as f:
    readme = f.read()

setup(
    name="serpyco",
    use_scm_version=True,
    description="Fast serialization of dataclasses using Cython",
    long_description=readme,
    author="Sébastien Grignard",
    author_email="pub@amakaze.org",
    url="https://gitlab.com/sgrignard/serpyco",
    packages=find_packages(),
    package_data={"serpyco": ["*.pyi", "py.typed"]},
    include_package_data=True,
    python_requires=">=3.6",
    setup_requires=[
        # Setuptools 18.0 properly handles Cython extensions.
        "setuptools>=18.0",
        "cython",
        "pytest-runner",
        "setuptools_scm",
        "wheel",
    ],
    install_requires=requires,
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries",
    ],
    ext_modules=[
        Extension("serpyco.serializer", sources=["serpyco/serializer.pyx"]),
        Extension("serpyco.encoder", sources=["serpyco/encoder.pyx"]),
    ],
    zip_safe=False,
)



# -*- coding: utf-8 -*-
# cython: language_level=3
# cython: embedsignature=True
# cython: wraparound=False
# cython: nonecheck=False
# cython: boundscheck=False


import dataclasses
import datetime
import enum
import re
import typing
import uuid

import cython
import dateutil.parser
import rapidjson
import typing_inspect

from serpyco.decorator import DecoratorType, _serpyco_tags

from serpyco.encoder cimport FieldEncoder

from serpyco.exception import NoEncoderError, NotADataClassError, ValidationError
from serpyco.field import FieldHints, _metadata_name
from serpyco.schema import SchemaBuilder
from serpyco.util import (
    JSON_ENCODABLE_TYPES,
    UNTYPED_DICT_TYPES,
    UNTYPED_ITERABLE_TYPES,
    JsonDict,
    JsonEncodable,
    _DataClassParams,
    _is_generic,
    _is_union,
    _issubclass_safe,
)
from serpyco.validator import RapidJsonValidator


cdef object dataclasses_MISSING = dataclasses.MISSING
cdef object new_object = object.__new__
_ITERABLE_TYPES_MAPPING = {
    typing.Tuple: tuple,
    typing.List: list,
    typing.Set: set,
}

cdef class SerializerField:
    cdef str field_name
    cdef str dict_key
    cdef FieldEncoder encoder
    cdef object getter
    cdef bint init
    cdef object default
    cdef object default_factory

    def __init__(
        self,
        str field_name,
        str dict_key,
        FieldEncoder encoder,
        object getter,
        bint init,
        object default,
        object default_factory
    ):
        self.field_name = field_name
        self.dict_key = dict_key
        self.encoder = encoder
        self.getter = getter
        self.init = init
        self.default = default
        self.default_factory = default_factory


cdef class Caster(object):
    cdef str dict_key
    cdef object caster

    def __cinit__(self, str dict_key, object caster):
        self.dict_key = dict_key
        self.caster = caster

cdef inline int cast_fields(tuple casters, dict data) except -1:
    cdef Caster caster
    for caster in casters:
        try:
            v = data[caster.dict_key]
        except KeyError:
            continue
        if _is_union(caster.caster):
            types = list(typing_inspect.get_args(caster.caster, evaluate=True))
            types.remove(type(None))
        else:
            types = [caster.caster]
        casted = None
        exc = None
        for type_ in types:
            try:
                casted = type_(v)
                break
            except Exception as ex:
                exc = ex
        if casted is None:
            raise ValidationError(f"Could not cast field {caster.dict_key}: {exc}")
        data[caster.dict_key] = casted


@cython.final
cdef class Serializer(object):
    """
    Serializer class for dataclasses instances.
    """

    cdef tuple _fields
    cdef object _dataclass_params
    cdef object _dataclass
    cdef bint _frozen_dataclass
    cdef bint _has_post_init
    cdef object _validator
    cdef list _parent_serializers
    cdef list _pre_dumpers
    cdef list _post_dumpers
    cdef list _pre_loaders
    cdef list _post_loaders
    cdef tuple _field_casters
    cdef dict _field_encoders
    cdef dict _type_encoders
    cdef object _load_as_type
    cdef list _only
    cdef tuple _excluded_fields
    _global_types = {
        datetime.datetime: DateTimeFieldEncoder(),
        uuid.UUID: UuidFieldEncoder()
    }
    for f, e in _global_types.items():
        SchemaBuilder.register_global_type(f, e)

    def __cinit__(
        self,
        dataclass,
        omit_none: bool = True,
        type_encoders: typing.Dict[type, FieldEncoder] = None,
        only: typing.Optional[typing.List[str]] = None,
        exclude: typing.Optional[typing.List[str]] = None,
        strict: bool = False,
        load_as_type: typing.Optional[type] = None,
        _parent_serializers: typing.List["Serializer"] = None
    ):
        """
        Constructs a serializer for the given data class.

        :param dataclass: data class this serializer will handle
        :param many: if True, serializer will handle lists of the dataclass
        :param omit_none: if False, keep None values in the serialized dicts
        :param type_encoders: encoders to use for given types
        :param only: list of fields to serialize.
            If None, all fields are serialized
        :param exclude: list of fields to exclude from serialization.
            If None, all fields are serialized
        :param strict: if true, unknown properties of an object will make the
            validation fail
        """
        cdef Serializer parent
        cdef SerializerField field
        self._dataclass_params = _DataClassParams(dataclass)
        self._dataclass = self._dataclass_params.type_
        self._type_encoders = type_encoders or {}
        self._parent_serializers = _parent_serializers or []
        self._parent_serializers.append(self)
        self._excluded_fields = tuple()
        fields = []
        excluded_fields = []
        field_casters = []
        type_hints = typing.get_type_hints(self._dataclass)
        self._field_encoders = {}
        for f in dataclasses.fields(self._dataclass):
            hints = f.metadata.get(_metadata_name, FieldHints(dict_key=f.name))
            if hints.dict_key is None:
                hints.dict_key = f.name

            field = SerializerField(
                f.name,
                hints.dict_key,
                None,
                hints.getter,
                f.init,
                f.default,
                f.default_factory
            )
            if (
                hints.ignore
                or (only and f.name not in only)
                or (exclude and f.name in exclude)
            ):
                excluded_fields.append(field)
            else:
                field_type = self._dataclass_params.resolve_type(type_hints[f.name])
                encoder = self._get_encoder(field_type, hints)
                if encoder:
                    self._field_encoders[field_type] = encoder
                    field.encoder = encoder
                if hints.cast_on_load:
                    field_casters.append(Caster(hints.dict_key, field_type))
                fields.append(field)

        self._fields = tuple(fields)
        self._excluded_fields = tuple(excluded_fields)

        field_encoders = {}
        for parent in self._parent_serializers:
            field_encoders.update(parent._field_encoders)
        builder = SchemaBuilder(
            dataclass,
            only=only,
            exclude=exclude,
            type_encoders={**self._global_types, **self._type_encoders},
            strict=strict,
        )
        self._validator = RapidJsonValidator(builder)

        # pre/post load/dump methods
        self._post_dumpers = []
        self._pre_dumpers = []
        self._post_loaders = []
        self._pre_loaders = []
        for attr_name in dir(self._dataclass):
            attr = getattr(self._dataclass, attr_name)
            try:
                tag = getattr(attr, _serpyco_tags)
                if DecoratorType.POST_DUMP==tag:
                    self._post_dumpers.append(attr)
                elif DecoratorType.PRE_DUMP==tag:
                    self._pre_dumpers.append(attr)
                elif DecoratorType.POST_LOAD==tag:
                    self._post_loaders.append(attr)
                elif DecoratorType.PRE_LOAD==tag:
                    self._pre_loaders.append(attr)
                else:
                    raise ValueError(f"Unknown decorator type {tag}")
            except AttributeError:
                continue
        self._field_casters = tuple(field_casters)
        self._load_as_type = load_as_type or self._dataclass
        # Check if the dataclass is frozen or not
        try:
            if self._fields:
                field = self._fields[0]
                obj = new_object(self._load_as_type)
                setattr(obj, field.field_name, None)
            self._frozen_dataclass = False
        except dataclasses.FrozenInstanceError:
            self._frozen_dataclass = True
        self._has_post_init = hasattr(self._dataclass, "__post_init__")

    def __hash__(self):
        cdef SerializerField sfield
        excluded_field_names = []
        for sfield in self._excluded_fields:
            excluded_field_names.append(sfield.field_name)
        return hash((
            self._dataclass,
            self._dataclass_params.arguments,
            tuple(excluded_field_names)
        ))

    def json_schema(self, many: bool = False) -> JsonDict:
        """
        Returns the JSON schema of the underlying validator.
        """
        return self._validator.json_schema(many=many)

    def get_dict_path(self, obj_path: typing.Sequence[str]) -> typing.List[str]:
        """
        Returns the path of a field in dumped dictionaries.
        :param obj_path: list of field names, for example
        ["foo", "bar"] to get the dict path of foo.bar
        """
        cdef SerializerField sfield
        cdef Serializer ser
        part = obj_path[0]
        for sfield in self._fields:
            if sfield.field_name==part:
                break
        else:
            raise KeyError(f"Unknown field {part} in {self._dataclass}")

        if 1 == len(obj_path):
            return [sfield.dict_key]

        ser = self._get_field_serializer(sfield)
        return [sfield.dict_key] + ser.get_dict_path(obj_path[1:])

    def get_object_path(self, dict_path: typing.Sequence[str]) -> typing.List[str]:
        """
        Returns the path of a field in loaded objects.
        :param dict_path: list of dictionary keys, for example
        ["foo", "bar"] to get the object path of {"foo": {"bar": 42}}
        """
        cdef SerializerField sfield
        cdef Serializer ser
        part = dict_path[0]
        for sfield in self._fields:
            if sfield.dict_key==part:
                break
        else:
            raise KeyError(f"Unknown dict key {part} in {self._dataclass}")

        if 1 == len(dict_path):
            return [sfield.field_name]

        ser = self._get_field_serializer(sfield)
        return [sfield.field_name] + ser.get_object_path(dict_path[1:])

    @classmethod
    def register_global_type(
        cls,
        field_type: typing.Any,
        encoder: FieldEncoder,
    ) -> None:
        """
        Registers a encoder/decoder for the given type.
        """
        cls._global_types[field_type] = encoder
        SchemaBuilder.register_global_type(field_type, encoder.json_schema())

    @classmethod
    def unregister_global_type(cls, field_type: typing.Any) -> None:
        """
        Removes a previously registered encoder for the given type.
        """
        del cls._global_types[field_type]
        SchemaBuilder.unregister_global_type(field_type)

    def dataclass(self) -> type:
        """
        Returns the dataclass used to construct this serializer.
        """
        return self._dataclass

    cpdef inline dump(
        self,
        obj: typing.Union[object, typing.Iterable[object]],
        bint validate: bool=False,
        bint many: bool=False,
    ):
        """
        Dumps the object(s) in the form of a dict/list only
        composed of builtin python types.

        :param validate: if True, the dumped data will be validated.
        """
        cdef list objs
        if many:
            objs = obj
            for pre_dump in self._pre_dumpers:
                objs = map(pre_dump, objs)
            data = [self._dump(o) for o in objs]
        else:
            for pre_dump in self._pre_dumpers:
                obj = pre_dump(obj)
            data = self._dump(obj)

        if validate:
            self._validator.validate(data, many=many)
            self._validator.validate_user(data, many=many)

        if many:
            for post_dump in self._post_dumpers:
                data = map(post_dump, data)
        else:
            for post_dump in self._post_dumpers:
                data = post_dump(data)

        return data

    cpdef inline load(
        self,
        data: typing.Union[dict, typing.Iterable[dict]],
        bint validate: bool=True,
        bint many: bool=False,
    ):
        """
        Loads the given data and returns object(s) of this serializer's
        dataclass.

        :param validate: if True, the data will be validated before
            creating objects
        """
        cdef list datas
        cdef list objs
        cdef object obj

        if many:
            datas = data
            if self._field_casters:
                for data in datas:
                    cast_fields(self._field_casters, data)
            for pre_load in self._pre_loaders:
                datas = map(pre_load, datas)
            data = datas
        else:
            if self._field_casters:
                cast_fields(self._field_casters, data)
            for pre_load in self._pre_loaders:
                data = pre_load(data)

        if validate:
            self._validator.validate(data, many=many)
            self._validator.validate_user(data, many=many)

        if many:
            objs = [self._load(d) for d in datas]
            for post_load in self._post_loaders:
                objs = map(post_load, objs)
            return objs

        obj = self._load(data)
        for post_load in self._post_loaders:
            obj = post_load(obj)
        return obj

    cpdef inline str dump_json(
        self,
        obj: typing.Union[object, typing.Iterable[object]],
        bint validate: bool=False,
        bint many: bool=False
    ):
        """
        Dumps the object(s) in the form of a JSON string.

        :param validate: if True, the dumped data will be validated
        """
        cdef list objs

        if many:
            objs = obj
            for pre_dump in self._pre_dumpers:
                objs = map(pre_dump, objs)
            data = [self._dump(o) for o in objs]
        else:
            for pre_dump in self._pre_dumpers:
                obj = pre_dump(obj)
            data = self._dump(obj)

        # Needed to validate
        js = rapidjson.dumps(data)

        if validate:
            self._validator.validate_json(js, many=many)
            self._validator.validate_user(data, many=many)

        if not self._post_dumpers:
            return js

        if many:
            for post_dump in self._post_dumpers:
                data = map(post_dump, data)
        else:
            for post_dump in self._post_dumpers:
                data = post_dump(data)

        # We need to dump in JSON again as post_dump can modify data.
        return rapidjson.dumps(data)

    cpdef inline load_json(
        self,
        str js,
        bint validate: bool=True,
        bint many: bool=False
    ):
        """
        Loads the given JSON string and returns object(s) of this serializer's
        dataclass.

        :param validate: if True, the JSON will be validated before
            creating objects
        """
        cdef list datas
        cdef list objs
        cdef object obj

        data = rapidjson.loads(js)

        if many:
            datas = data
            if self._field_casters:
                for data in datas:
                    cast_fields(self._field_casters, data)
            for pre_load in self._pre_loaders:
                datas = map(pre_load, datas)
            data = datas
        else:
            if self._field_casters:
                cast_fields(self._field_casters, data)
            for pre_load in self._pre_loaders:
                data = pre_load(data)

        if validate:
            self._validator.validate(data, many=many)
            self._validator.validate_user(data, many=many)

        if many:
            objs = [self._load(d) for d in datas]
            for post_load in self._post_loaders:
                objs = map(post_load, objs)
            return objs
        obj = self._load(data)
        for post_load in self._post_loaders:
            obj = post_load(obj)
        return obj

    cdef inline dict _dump(self, object obj):
        cdef SerializerField sfield
        cdef object encoded
        cdef dict data = {}
        for sfield in self._fields:
            if sfield.getter is not None:
                encoded = sfield.getter(obj)
            else:
                encoded = getattr(obj, sfield.field_name)
            if sfield.encoder is not None and encoded is not None:
                encoded = sfield.encoder.dump(encoded)
            data[sfield.dict_key] = encoded
        return data

    cdef inline object _load(self, dict data):
        cdef SerializerField sfield
        cdef object decoded
        cdef object obj
        obj = new_object(self._load_as_type)
        for sfield in self._fields:
            try:
                decoded = data[sfield.dict_key]
                if sfield.encoder is not None and decoded is not None:
                    decoded = sfield.encoder.load(decoded)
            except KeyError:
                if sfield.default is not dataclasses_MISSING:
                    decoded = sfield.default
                elif sfield.default_factory is not dataclasses_MISSING:
                    decoded = sfield.default_factory()
                else:
                    raise TypeError(
                        "data dictionary is missing "
                        f"required parameter '{sfield.field_name}' "
                        f"for class '{self._dataclass.__qualname__}'"
                    )
            # Cannot use setattr() on frozen dataclasses
            if not self._frozen_dataclass:
                setattr(obj, sfield.field_name, decoded)
            else:
                object.__setattr__(obj, sfield.field_name, decoded)
        for sfield in self._excluded_fields:
            if sfield.default is not dataclasses.MISSING:
                decoded = sfield.default
            elif sfield.default_factory is not dataclasses.MISSING:
                decoded = sfield.default_factory()
            else:
                raise TypeError(
                    f"Cannot create '{self._dataclass.__qualname__}' as "
                    f"'{sfield.field_name}' is excluded but has no "
                    "default value/factory"
                )
            # Cannot use setattr() on frozen dataclasses
            if not self._frozen_dataclass:
                setattr(obj, sfield.field_name, decoded)
            else:
                object.__setattr__(obj, sfield.field_name, decoded)
        if self._has_post_init:
            obj.__post_init__()
        return obj

    def _get_field_serializer(self, sfield: SerializerField) -> "Serializer":
        cdef FieldEncoder encoder = sfield.encoder
        cdef DataClassFieldEncoder dencoder
        cdef IterableFieldEncoder iter_encoder
        cdef DataClassIterableFieldEncoder diter_encoder
        cdef DictFieldEncoder dict_encoder
        encoder = sfield.encoder
        if isinstance(encoder, IterableFieldEncoder):
            iter_encoder = encoder
            dencoder = iter_encoder._item_encoder
            return dencoder._serializer
        elif isinstance(encoder, DataClassIterableFieldEncoder):
            diter_encoder = encoder
            return diter_encoder._serializer
        elif isinstance(encoder, DictFieldEncoder):
            dict_encoder = encoder
            dencoder = dict_encoder._value_encoder
            return dencoder._serializer
        elif isinstance(encoder, DataClassFieldEncoder):
            dencoder = encoder
            return dencoder._serializer
        raise ValueError(f"field {sfield.field_name} is not a dataclass")

    def _get_encoder(self, field_type, hints):
        cdef DataClassFieldEncoder dencoder
        field_type = self._dataclass_params.resolve_type(field_type)
        args = typing_inspect.get_args(field_type, evaluate=True)

        if typing_inspect.is_new_type(field_type):
            field_type = field_type.__supertype__

        if field_type in self._type_encoders:
            return self._type_encoders[field_type]
        elif field_type in self._global_types:
            return self._global_types[field_type]
        elif typing.Any == field_type:
            return None
        elif _issubclass_safe(field_type, enum.Enum):
            # Must be first as enums can inherit from another type
            return EnumFieldEncoder(field_type)
        elif _issubclass_safe(field_type, tuple(JSON_ENCODABLE_TYPES.keys())):
            return None
        elif _is_union(field_type):
            args = list(typing_inspect.get_args(field_type, evaluate=True))
            try:
                args.remove(type(None))
            except ValueError:
                pass
            type_encoders = [
                (item_type, self._get_encoder(item_type, hints))
                for item_type in args
            ]
            if not args:
                return None
            elif len(args)== 1:
                return type_encoders[0][1]
            else:
                return UnionFieldEncoder(type_encoders)
        elif field_type in UNTYPED_DICT_TYPES:
            return None
        elif _is_generic(field_type, typing.Mapping):
            key_encoder = self._get_encoder(args[0], hints)
            value_encoder = self._get_encoder(args[1], hints)
            if key_encoder or value_encoder:
                return DictFieldEncoder(key_encoder, value_encoder)
            return None
        elif (
            _is_generic(field_type, typing.Tuple)
            and (
                len(args)!=2
                or args[len(args)-1] is not ...
            )
        ):
            # Special case for tuple with fixed-length argument list
            item_encoders = [
                self._get_encoder(arg_type, hints)
                for arg_type in args
            ]
            return FixedTupleFieldEncoder(item_encoders)
        elif field_type in UNTYPED_ITERABLE_TYPES:
            return IterableFieldEncoder(None, field_type)
        # tuples defined with ... are handled by the following elif
        elif _is_generic(field_type, typing.Iterable):
            item_encoder = self._get_encoder(args[0], hints)
            if isinstance(item_encoder, DataClassFieldEncoder):
                dencoder = item_encoder
                return DataClassIterableFieldEncoder(dencoder._serializer, field_type)
            return IterableFieldEncoder(item_encoder, field_type)
        # Is the field a dataclass ?
        try:
            params = _DataClassParams(field_type)
        except NotADataClassError:
            raise NoEncoderError(f"No encoder for '{field_type}'")

        # See if one of our "ancestors" handles this dataclass.
        # This avoids infinite recursion if dataclasses establish a cycle
        for serializer in self._parent_serializers:
            excluded_field_names = tuple(
                f.name for f in dataclasses.fields(params.type_)
                if (
                    hints.ignore
                    or (hints.only and f.name not in hints.only)
                    or (hints.exclude and f.name in hints.exclude)
                )
            )
            sh = hash(serializer)
            h = hash((params.type_, params.arguments, excluded_field_names))
            if h == sh:
                break
        else:
            serializer = Serializer(
                field_type,
                type_encoders=hints.type_encoders or self._type_encoders,
                only=hints.only,
                exclude=hints.exclude,
                load_as_type=hints.load_as_type,
                _parent_serializers=self._parent_serializers,
            )
        return DataClassFieldEncoder(serializer)


# builtin field encoders
@cython.final
cdef class EnumFieldEncoder(FieldEncoder):
    cdef object _enum_type

    def __cinit__(self, enum_type):
        self._enum_type = enum_type

    cpdef inline dump(self, value: typing.Any):
        return value.value

    cpdef inline load(self, value: typing.Any):
        return self._enum_type(value)

    def json_schema(self) -> JsonDict:
        return None


@cython.final
cdef class DataClassFieldEncoder(FieldEncoder):
    cdef Serializer _serializer

    def __cinit__(self, Serializer serializer):
        self._serializer = serializer

    cpdef inline load(self, value):
        return self._serializer._load(value)

    cpdef inline dump(self, value):
        return self._serializer._dump(value)

    def json_schema(self) -> JsonDict:
        return None


@cython.final
cdef class DataClassIterableFieldEncoder(FieldEncoder):
    cdef Serializer _serializer
    cdef object _iterable_type

    def __cinit__(self, serializer, sequence_type):
        origin = typing_inspect.get_origin(sequence_type)
        self._iterable_type = _ITERABLE_TYPES_MAPPING.get(origin, origin)
        self._serializer = serializer

    cpdef inline load(self, value):
        cdef lst = [
            self._serializer._load(v)
            if v is not None
            else None
            for v in value
        ]
        return self._iterable_type(lst)

    cpdef inline dump(self, value):
        return [
            self._serializer._dump(v)
            if v is not None
            else None
            for v in value
        ]

    def json_schema(self) -> JsonDict:
        return None


@cython.final
cdef class FixedTupleFieldEncoder(FieldEncoder):
    cdef tuple _item_encoders
    cdef int _item_encoders_count

    def __cinit__(self, item_encoders):
        self._item_encoders = tuple(item_encoders)
        self._item_encoders_count = len(self._item_encoders)

    cpdef inline load(self, value: typing.Any):
        cdef FieldEncoder encoder
        cdef int i
        decoded = []
        if len(value) != self._item_encoders_count:
            raise ValidationError("Invalid number of items for tuple")
        for i in range(0, self._item_encoders_count):
            encoder = self._item_encoders[i]
            if encoder:
                decoded.append(encoder.load(value[i]))
            else:
                decoded.append(value[i])
        return tuple(decoded)

    cpdef inline dump(self, value: typing.Any):
        cdef FieldEncoder encoder
        cdef int i
        encoded = []
        if len(value) != self._item_encoders_count:
            raise ValidationError("Invalid number of items for tuple")
        for i in range(0, self._item_encoders_count):
            encoder = self._item_encoders[i]
            if encoder:
                encoded.append(encoder.dump(value[i]))
            else:
                encoded.append(value[i])
        return encoded

    def json_schema(self) -> JsonDict:
        return None


@cython.final
cdef class IterableFieldEncoder(FieldEncoder):
    cdef FieldEncoder _item_encoder
    cdef object _iterable_type

    def __cinit__(self, item_encoder, sequence_type):
        self._item_encoder = item_encoder
        origin = typing_inspect.get_origin(sequence_type)
        self._iterable_type = _ITERABLE_TYPES_MAPPING.get(origin, origin)

    cpdef inline load(self, value: typing.Any):
        if self._item_encoder:
            value = [
                self._item_encoder.load(v) if v is not None
                else None
                for v in value
            ]
        return self._iterable_type(value)

    cpdef inline dump(self, value: typing.Any):
        if self._item_encoder:
            return [
                self._item_encoder.dump(v) if v is not None
                else None
                for v in value
            ]
        return list(value)

    def json_schema(self) -> JsonDict:
        return None


@cython.final
cdef class DictFieldEncoder(FieldEncoder):
    cdef FieldEncoder _key_encoder
    cdef FieldEncoder _value_encoder

    def __cinit__(self, key_encoder, value_encoder):
        self._key_encoder = key_encoder
        self._value_encoder = value_encoder

    cpdef inline load(self, value: JsonEncodable):
        if self._key_encoder and self._value_encoder:
            return {
                self._key_encoder.load(k): self._value_encoder.load(v)
                for k, v in value.items()
            }
        elif self._key_encoder and not self._value_encoder:
            return {
                self._key_encoder.load(k): v
                for k, v in value.items()
            }
        elif not self._key_encoder and self._value_encoder:
            return {
                k: self._value_encoder.load(v)
                for k, v in value.items()
            }
        else:
            return value

    cpdef inline dump(self, value: typing.Any):
        if self._key_encoder and self._value_encoder:
            return {
                self._key_encoder.dump(k): self._value_encoder.dump(v)
                for k, v in value.items()
            }
        elif self._key_encoder and not self._value_encoder:
            return {
                self._key_encoder.dump(k): v
                for k, v in value.items()
            }
        elif not self._key_encoder and self._value_encoder:
            return {
                k: self._value_encoder.dump(v)
                for k, v in value.items()
            }
        else:
            return value

    def json_schema(self) -> JsonDict:
        return None


@cython.final
cdef class DateTimeFieldEncoder(FieldEncoder):
    """Encodes datetimes to RFC3339/ISO8601 format"""

    iso8601_pattern = (
        r"^[0-9]{4}-[0-9][0-9]-[0-9][0-9]T"  # YYYY-MM-DD
        r"[0-9][0-9]:[0-9][0-9]:[0-9][0-9](\.[0-9]+)"  # HH:mm:ss.ssss
        r"?(([+-][0-9][0-9]:[0-9][0-9])|Z)?$"  # timezone
    )

    cpdef inline dump(self, value):
        try:
            return value.isoformat()
        except AttributeError:
            raise ValidationError(f"{value} is not a datetime.datetime instance")

    cpdef inline load(self, value):
        try:
            return dateutil.parser.parse(value)
        except (ValueError, OverflowError):
            raise ValidationError(f"{value} is not a valid datetime")

    def json_schema(self) -> JsonDict:
        return {
            "type": "string",
            "format": "date-time",
            "pattern": self.iso8601_pattern
        }


@cython.final
cdef class UuidFieldEncoder(FieldEncoder):

    cpdef inline dump(self, value):
        return str(value)

    cpdef inline load(self, value):
        return uuid.UUID(value)

    def json_schema(self) -> JsonDict:
        return {"type": "string", "format": "uuid"}


@cython.final
cdef class UnionFieldEncoder(FieldEncoder):

    cdef tuple _type_encoders

    def __cinit__(
        self,
        type_encoders: typing.List[typing.Tuple[type, FieldEncoder]]
    ):
        self._type_encoders = tuple(type_encoders)

    cpdef inline dump(self, value):
        cdef list value_types = []
        for value_type, encoder in self._type_encoders:
            value_types.append(str(value_type))
            try:
                return encoder.dump(value) if encoder else value
            except Exception:
                pass
        union = ",".join(value_types)
        msg = f"{value} has a wrong type, expected any of [{union}]"
        raise ValidationError(msg)

    cpdef inline load(self, value):
        cdef list value_types = []
        for value_type, encoder in self._type_encoders:
            try:
                value_types.append(str(value_type))
                return encoder.load(value) if encoder else value
            except Exception:
                pass
        union = ",".join(value_types)
        msg = f"{value} has a wrong type, expected any of [{union}]"
        raise ValidationError(msg)

    def json_schema(self) -> JsonDict:
        return None



import typing

from serpyco.serializer import Serializer

T = typing.TypeVar("T", bound="SerializerMixin")


class SerializerMixin:
    """Base class that provides load/dump, load_json/dump_json methods."""

    _serializer: typing.ClassVar[typing.Optional[Serializer]] = None  # type: ignore

    @classmethod
    def load(
        cls: typing.Type[T], data: typing.Dict[str, typing.Any], validate: bool = True
    ) -> T:
        """Load the given dict an return a new object."""
        return typing.cast(T, cls.serializer().load(data, validate=validate))

    def dump(self, validate: bool = False) -> typing.Dict[str, typing.Any]:
        """Dump the object to a dict."""
        return typing.cast(
            typing.Dict[str, typing.Any],
            self.serializer().dump(self, validate=validate),
        )

    @classmethod
    def load_json(cls: typing.Type[T], json_string: str, validate: bool = True) -> T:
        """Load the given JSON string an return a new object."""
        return typing.cast(
            T, cls.serializer().load_json(json_string, validate=validate)
        )

    def dump_json(self, validate: bool = False) -> str:
        """Dump the object to a JSON string."""
        return self.serializer().dump_json(self, validate=validate)

    @classmethod
    def serializer(cls: typing.Type[T]) -> Serializer:  # type: ignore
        """Serializer instance for this class."""
        if not cls._serializer:
            cls._serializer = Serializer(cls)
        return cls._serializer
